__version__ = "0.2.6"
__author__ = "Wilhelm Lab"
__author_email__ = "o.shouman@tum.de"
__license__ = "MIT"
__description__ = "Deep Learning for Proteomics"
__package__ = "DLOmix"
__copyright__ = "2026, Wilhelm Lab, TU Munich, School of Life Sciences"
__github_url__ = "https://github.com/wilhelm-lab/dlomix"
